import asyncio
import logging
import pickle
import struct
from typing import Tuple

from aiocarbon.metric import Metric
from .base import BaseClient, chunk_list

log = logging.getLogger(__name__)


class PickleClient(BaseClient):
    CHUNK_SIZE = 512

    async def send(self):
        async with self as metrics:
            reader, writer = await asyncio.open_connection(
                self._host, self._port, loop=self.loop
            )

            for chunk in chunk_list(metrics, self.CHUNK_SIZE):
                payload = pickle.dumps(
                    [self.format_metric(m) for m in chunk],
                    protocol=2
                )

                header = struct.pack("!L", len(payload))
                writer.write(header)
                await writer.drain()

                writer.write(payload)

            await writer.drain()
            writer.close()
            reader.feed_eof()

    def format_metric(self, metric: Metric) -> Tuple[str, Tuple[float, int]]:
        return (
            self.format_metric_name(metric),
            (metric.timestamp, metric.value)
        )



import abc
from collections import defaultdict
from typing import Optional

import time

from .protocol.base import BaseClient
from .metric import Metric


class MeasurementBase:
    CLIENT = None       # type: BaseClient

    __slots__ = "_name",

    def __init__(self, name):
        self._name = name

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        metric = self.get_metric()

        if metric:
            self.CLIENT.add(metric)

    @abc.abstractmethod
    def get_metric(self) -> Optional[Metric]:
        raise NotImplementedError


class Meter(MeasurementBase):
    VALUES = defaultdict(lambda: defaultdict(int))

    def get_metric(self) -> Metric:
        return Metric(self._name, self.value)

    @property
    def value(self):
        return self.VALUES[self.__class__][self._name]

    @value.setter
    def value(self, value):
        self.VALUES[self.__class__][self._name] = value

    def set(self, value):
        self.value = value


class Counter(MeasurementBase):
    COUNTERS = defaultdict(lambda: defaultdict(int))

    def get_metric(self) -> Optional[Metric]:
        return None

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            name = "%s.%s" % (self._name.rstrip("."), 'fail')
        else:
            name = "%s.%s" % (self._name.rstrip("."), 'ok')

        self.COUNTERS[self.__class__][name] += 1
        value = self.COUNTERS[self.__class__][name]

        self.CLIENT.add(Metric(name, value))


class Timer(MeasurementBase):
    TIMERS = defaultdict(lambda: defaultdict(int))

    def get_metric(self) -> Optional[Metric]:
        return None

    def __enter__(self):
        self._start_time = time.monotonic()
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        delta = time.monotonic() - self._start_time

        if exc_type:
            name = "%s.%s" % (self._name.rstrip("."), 'fail')
        else:
            name = "%s.%s" % (self._name.rstrip("."), 'ok')

        self.CLIENT.add(Metric(name, delta))


def set_client(client: BaseClient):
    MeasurementBase.CLIENT = client


__all__ = "Meter", "Counter", "Timer", "set_client",



import abc
import asyncio
import logging
from collections import deque
from typing import Iterable, Any

import itertools

from aiocarbon.metric import Metric


log = logging.getLogger(__name__)


class BaseClient:
    SEND_PERIOD = 1

    def __init__(self, host: str, port: int = 2003,
                 loop: asyncio.AbstractEventLoop = None,
                 namespace: str=None):
        self.loop = loop or asyncio.get_event_loop()
        self._ns = namespace if namespace is not None else 'carbon'
        self._host = host
        self._port = port
        self._metrics = deque()
        self._lock = asyncio.Lock(loop=self.loop)
        self.__metrics = deque()

    async def run(self):
        while True:
            try:
                await self.send()
            except:
                log.exception("Exception while sending metrics")
            finally:
                await asyncio.sleep(self.SEND_PERIOD, loop=self.loop)

    async def __aenter__(self):
        await self._lock.acquire()
        while self._metrics:
            metric = self._metrics.popleft()
            self.__metrics.append(metric)

        return self.__metrics

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            while self.__metrics:
                self._metrics.appendleft(self.__metrics.pop())
        else:
            self.__metrics.clear()

        self._lock.release()

    def format_metric_name(self, metric: Metric):
        if self._ns:
            return ".".join((self._ns, metric.name))
        else:
            return metric.name

    @abc.abstractmethod
    def send(self):
        raise NotImplementedError

    @abc.abstractmethod
    def format_metric(self, metric: Metric):
        raise NotImplementedError

    def add(self, metric: Metric):
        self._metrics.append(metric)


def chunk_list(iterable: Iterable[Any], size: int):
    iterable = iter(iterable)

    item = list(itertools.islice(iterable, size))

    while item:
        yield item
        item = list(itertools.islice(iterable, size))
